/* dCGISession A class that mangages CGI operations when dBASE is used  to
					create Web Applets. The current version supports only CGI.
               Future versions will support ISAPI and NSAPI and a likely
               dBASE application server.

   Usage:		oCGI = new CGISession()

	Author:		A.A.Katz, author
               v.1.0 11/27/1999
               v.1.1 01/12/2000 (supports win 98 StdOut)
               v.1.2 05/24/2001 Loop with timeout added to STDIN read
                                __asian__ build OEMFormat method added
               v.1.3 08/24/2001 loadArrayFromFields method revised
               v.1.4 09/04/2001 Chinese, Korean added to OEMFormat method #ifdef __asian
               v.1.5 01/31/2002 'AUTOINC' string comparisons standardized

   Rights:		You have the right to freely include this code in any compiled
   				dBASE Web application. You do not have the right to reproduce
               or publish this source code without the explicit written
               permission of dBASE, Inc., Vestal, NY
               (c) 1999-2000 dBASE Inc.
   Platform:	This code works only with Visual dBASE 7.5 or later.
   				If this code is used, you do not need to use VdB_CGI.exe
               as required with earlier versions. Your Visual dBASE
               applications may be called directly by the Web Server.
*/


Class CGISession of assocArray
   ////// Constructor code

   this.nTimeout = 45             // STDIN read timeout in seconds #####

   ////// Set dBASE environment commands. Web apps are non-GUI

	SET CENTURY ON
   SET TALK OFF




   ////// Method:		Connect //////////////////////////////////////////////////
   ////// Purpose		Connects by StdIn and StdOut to Web server ///////////////
   //////            reads in data, and then parses out name/value pairs //////
   //////				and creates a new element of this array for each. ////////

   Function Connect

      // Set default WebMaster EMail address

      this["WebMasterEMail"] = ""

      // Read incoming data into this array as name/value pairs
      // Ex: this["City] = "Vestal", where "City" is the name
      // and "Vestal" is the value

      this.loadArrayFromCGI()

      ////// Open  output stream back to Web Server
      Try

         this.fOut = new file()
         this.fOut.Open("StdOut", "RA")  // open the StdOut output stream

      Catch (exception e)

         Quit 							// Might as well quit, can't send response

      endTry


   ////// Method:  LoadArrayFromGGI //////////////////////////////////////
   ////// Purpose: Loads existing assoc array with data from  ////////////
   //////  			 Web Server from a previous page////////////////////////
   ////// Params:  None //////////////////////////////////////////////////

   /* Note: This code connects to the Web Server and captures the incoming
   			data sent when the HTML form was submitted. This Associative
            Array, fortunately, mimics the Web's Name/Value pairs.
            Example: this["City"] = "Vestal", where "City" is the name and
           "Vestal" is the value.
   */

   Function LoadArrayFromCGI

      // First, look in environment to see if the page was submitted via
      // Post or Get

      cMethod = upper(getEnv("REQUEST_METHOD"))

      if cMethod = 'POST'    // if the submit method was "Post" (Recommended!!!)
                             // get length of incoming data string
         nLen = val(getEnv("CONTENT_LENGTH"))
         nCnt = 0            // Initialize bytes read counter #####

         Try

            fIn  = new file()        // Create new instance of File Object
            fIn.Open("StdIn", "RA")  // open the StdIn input stream
            cEnv = ""                // Initialize read buffer
            cStart = time()          // Save start time
            // Loop until all bytes are read or timeout occurs
            do while nCnt < nLen
               cEnv += fIn.Read(nLen - nCnt)  // Read additional bytes
               nCnt := len ( cEnv )  // Update byte counter
               // Compute seconds elapsed, compensated for date rollover
               if this.nTimeout < IIf(elapsed(time(), cStart) < 0,;
                                      elapsed(time(), cStart) + 86400,;
                                      elapsed(time(), cStart))
                  exit               // Exit if timeout value exceeded
               endIf
            endDo

         Catch (exception e)

            Quit        				// Might as well quit, can't open connection

         endTry

         // Note: StdIn and StdOut cannot be close()'ed.

      else

         cEnv = getEnv("QUERY_STRING") // Get data from environment

      endif

      /* Note: Name/Value pairs are sent in a form something like this:
             	City=Vestal&State=NY&Country=USA  with "&" as delimeter
               between named pairs.

         Note: Punctuation and spaces are sent using special characters.
         		Furthermore, the Internet uses the International character
               standard: ANSI while dBASE uses its own internal character
               standard: OEM. The OEMFormat() method handles all
               conversions to dBASE-usable character strings automatically.
      */

      ////// Break up CGI input string into
      ////// Name/Value pairs

      try

         do while at("=",cEnv) > 0   // Do while there are still pairs

            cName  = ''             // init name/value vars
            cValue = ''


            If at("&",cEnv) > 0      // If there is more than one pair
                                     // Grab the pair before first "&'
               cEnvPair = this.OEMFormat(substr(cEnv,1,at("&",cEnv)-1))

                                     // Strip this pair off of cEnv
                                     // in preparation for next loop
               cEnv = substr(cEnv,at("&",cEnv)+1)

            else                     // This must be the last pair

               cEnvPair = this.OEMFormat(cEnv) // grab remainder
               cEnv = ''             // Nothing left to parse

            endif

            ////// separate name and value
            cName = substr(cEnvPair,1,at('=',cEnvPair)-1)+""
            cValue = substr(cEnvPair, at('=',cEnvPair)+1)+""

            ////// if name/value pair exists, append into this array
            if len(trim(cName)) >0 //

               this[cName] = cValue  // update or add to array

            endif

         enddo

      catch (exception e)

         this.errorPage(e)

      endtry

      return true


   ////// Method 		OEMFormat /////////////////////////////////////////////////
   ////// Purpose 	Transforms CGI escape chars to OEM/////////////////////////
   ////// Param: 		cStr 	Character strings sent by Web browser ///////////////
   //////			   in ANSI format ////////////////////////////////////////////

   /* Note: Strings come in over the internet in ANSI format and must be
            converted to dBASE OEM Format before use. Punctuation is
            sent in %nn hexadecimal format and spaces to plus signs (+).
            This method does all the conversion necessary to convert from
            ANSI CGI to dBASE OEM.
            Example: Alan's New Class, in ANSI CGI is
                     Alan%27s+New+Class
   */

   #ifdef __asian__

   Function OEMFormat(cStr) Protect // For internal use only
      // this version [May.2000] supports Shift-JIS double-byte Japanese

      // remove "Pluses" used as space keepers in CGI
      do while at("+",cStr) > 0
         cStr = substr(cStr,1,at("+",cStr)-1) + " "+;
                substr(cStr,at("+",cStr)+1)
      enddo

      // add in "CRs" sent as escape characters
      do while at("%0D%0A",cStr) > 0
         cStr = substr(cStr,1,at("%0D%0A",cStr)-1) +chr(13)+;
                substr(cStr,at("%0D%0A",cStr)+3)
      enddo

      // convert escape characters back to oem chars
      // non-ascii and Asian chars likely encoded as '%nn' per each byte
      // loop as long as '%' exists in string

      nLeft = 1
      do while at("%",substr(cStr,nLeft) ) > 0
         nPos = at("%",substr(cStr,nLeft))+nLeft-1

         nLeft = nPos+1
         if nPos <= len(cStr)-2

           jumpAhead = 3
           // convertincoming encoding to character
           char1 = chr( htoi(substr(cStr,nPos+1,2)) )
           validCharacter = true

           // is char1 a lead byte of a double-byte character?
           // Japanese Shift-JIS lead bytes are only in the following range

           // convert nn to decimal
           chrCode = htoi(substr(cStr,nPos+1,2))

           isLeadByte = false
           if '932' $ _app.charset // Japanese Shift-JIS
                                   // lead-byte range (non-Contiguous)
                                   //  81h - 9Fh and then E0h - FCh
              isLeadByte = ( chrCode >= 129 .and. chrCode <= 159 ) .or.;
                           ( chrCode >= 224 .and. chrCode <= 239 )

           elseif '936' $ _app.charset //  Simplified Chinese
                                  //  lead byte range A1h - FEh
              isLeadByte = ( chrCode >= 161 .and. chrCode <= 254 )

           elseif '949' $ _app.charset // Korean
                                  // leadByte range 81h - FEh
              isLeadByte = ( chrCode >= 129 .and. chrCode <= 254 )

           elseif '950' $ _app.charset // Traditional Chinese
                                  // lead byte range A1h - FEh
              isLeadByte = ( chrCode >= 161 .and. chrCode <= 254 )

           endif

           if isLeadByte
              // we have a lead byte
              // must grab the trail byte, and then convert to character

              // trail byte is not _always_ %nn encoded
              //   as CGI passes them thru as is,
              //   when they happen to match a lower-case ASCII codePoint
              if substr( cStr, nPos+3, 1 ) = "%"
                 // trail byte is %nn encoded
                 char1 = chr( htoi( substr(cStr,nPos+1,2) + substr( cStr, nPos+4, 2 )))
                 jumpAhead = 6
              else
                 // trail byte is plain text
                 char1 = chr( htoi( substr(cStr,nPos+1,2) + itoh( asc( substr( cStr, nPos+3, 1 ))) ))
                 jumpAhead = 4
              endif
              validCharacter = iskanji( char1 )
           endif

           // throw away character if invalid
           if validCharacter
              cStr = left(cStr,nPos-1)+ char1 +;
                  substr(cStr,nPos+jumpAhead)
           else
              cStr = left(cStr,nPos-1)+ substr(cStr,nPos+3)
           endif
         endif
      enddo

      return cStr

   #else

   Function OEMFormat(cStr) Protect // For internal use only

      // remove "Pluses" used as space keepers in CGI
      do while at("+",cStr) > 0
         cStr = substr(cStr,1,at("+",cStr)-1) + " "+;
                substr(cStr,at("+",cStr)+1)
      enddo

      // add in "CRs" sent as escape characters
      do while at("%0D%0A",cStr) > 0
         cStr = substr(cStr,1,at("%0D%0A",cStr)-1) +chr(13)+;
                substr(cStr,at("%0D%0A",cStr)+3)
      enddo

      // convert escape characters back to oem chars
      nLeft = 1
      do while at("%",substr(cStr,nLeft) ) > 0
         nPos = at("%",substr(cStr,nLeft))+nLeft-1
         nLeft = nPos+1
         if nPos <= len(cStr)-2
            cStr = left(cStr,nPos-1)+ chr(htoi(substr(cStr,nPos+1,2)))+;
            substr(cStr,nPos+3)
         endif
      enddo

      return cStr

   #endif



   ////// Method: 	SetWebMasterAddress //////////////////////////////////
   ////// Purpose:	Sets mailto: address used in Error page //////////////
   ////// Param:		cWebMasterEMail (ex: webmaster@my.server) ////////////

   Function SetWebMasterAddress(cMailAddress)
      if not empty(cMailAddress)
         this["WebMasterEMail"] = cMailAddress
      endif
      return true



   /////////////////////// Output Methods //////////////////////////////////



   ////// Method:   SorryPage //////////////////////////////////////////////
   ////// Purpose:    Returns Generic Sorry response page //////////////////
   ////// Param:      cMsg  = Error Message ////////////////////////////////
   //////             		   Accepts a char string or array of messages ///
   //////             cSubTtl Subtitle, such as 'An error has ocurred"  or /
   //////                    "data missing" ////////////////////////////////
   //////             cRecover = Suggestion for recovery ///////////////////
   //////                     such as "Press the browser's back button" ////
   //////                     or contact Web master.... ////////////////////

   Function SorryPage(cMsg,cSubTtl,cRcvr)

      aMsg = iif(empty(cMsg),'',cMsg)

      if type('aMsg') = 'C'   // if message is not array,
         aMsg = new array()   // convert to a one-element array
         aMsg.add(cMsg)
      endif
                               // Make private versions for type()
      cRecover =  iif(empty(cRcvr),'Press Back button and try again.',cRcvr)
      cSubtitle = iif(empty(cSubTtl),'',cSubTtl)


      /////// Stream out header
      this.StreamHeader('Sorry!')

      with (this.fOut)

         ////// Body tag (starts body of page)
         puts('<BODY>')
         puts('')
         puts('<P>')
         puts('<TABLE BORDER="0" CELLPADDING="0" CELLSPACING="0" WIDTH="100%">')

         ////// Sorry!
         puts('	<TR>')
         puts('		<TD COLSPAN="2" BGCOLOR="#E4E4E4">')
         puts('			<H1><FONT COLOR="#999999">Sorry!</FONT>')
         puts('		</TD>')
         puts('	</TR>')

         ////// Blank Line
         puts('	<TR>')
         puts('		<TD WIDTH="9%">&nbsp;</TD>')
         puts('		<TD WIDTH="91%">&nbsp;</TD>')
         puts('	</TR>')

      ////// Subtitle

         if len(trim(cSubTitle)) > 0 // if there's a subtitle

            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%"><I><FONT SIZE="3" FACE="Arial, Helvetica">'+cSubTitle+'</FONT></I></TD>')
            puts('	</TR>')
            puts('	<TR>')

            ////// Blank Line
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%">&nbsp;</TD>')
            puts('	</TR>')


         endif

      ////// Traverse the message array
      	for n = 1 to aLen(aMsg,1)

            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%"><FONT SIZE="2" FACE="Arial, Helvetica">'+aMsg[n]+'</FONT></TD>')
            puts('	</TR>')

         next

         ////// If recovery instructions exist (Contact Webmaster, Press Back
         ////// 		Button, etc.), stream it here
         if len(trim(cRecover)) > 0
            with (this.fOut)
            ////// Blank line
            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%">&nbsp;</TD>')
            puts('	</TR>')

            ////// Recovery message
            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%"><FONT SIZE="2" FACE="Arial, Helvetica">'+cRecover+'</FONT></TD>')
            puts('	</TR>')
            endwith
         endif


         puts('</TABLE>')
         puts('')
         puts('')
         puts('</BODY>')


      endwith

      ////// Close out HTML structure
      this.StreamFooter()

      quit


   /////// Method:	errorPage  ////////////////////////////////////////////////
   /////// Purpose:	Reports error back to user and quits application //////////
   /////// Params:	oErr - takes the error object thrown by the dBASE /////////
   ///////				error-handling system. Can be thrown by the developer, ////
   ///////				or passed from the "Catch" of a try...catch statement /////
   /////// Calls:		this.sorryPage to stream error page back to Server ////////

   Function errorPage(oErr)

      aError = new array()                // Create array for messages

      with (aError)
         add('File: '+oErr.Filename)     // Add file/line and message
         add('Line No:'+oErr.LineNo)
         add(''+oErr.Message)
      endwith

      // If WebMaster eMail address exists, use it as a Mailto: address
      // on the bottom of the error page.


      if len(this["WebMasterEMail"]) > 0
         cRecover  = '<I><FONT FACE="Arial, Helvetica">Please contact the </FONT>'+;
                     '</I><A HREF="mailto:'+this["WebMasterEMail"]+'"><I>'+;
                     '<FONT FACE="Arial, Helvetica">Webmaster</FONT></I></A>'+;
                     '<I><FONT FACE="Arial, Helvetica">.</FONT></I>'
      else
         cRecover = 'Contact the WebMaster'
      endif

                                          // Call sorryPage() to display
      this.sorryPage(aError,'An error ocurred on the server',cRecover)

      return true



   ////// Method: 	StreamHeader //////////////////////////////////////////////
   ////// Purpose		Streams the CGI and HTML headers required  ////////////////
   //////				to send a response page back through the server ///////////
   ////// Param:		Title - the page title desired ////////////////////////////

   Function StreamHeader(Title)
      cTitle = iif(empty(Title),'Response Page',Title)

      with (this.fOut)
         puts('Content-type: text/html')
         puts('')
         puts('<HTML>')
         puts('')
         puts('<HEAD>')
         puts('	<META HTTP-EQUIV="Content-Type" CONTENT="text/html">')
         puts('	<TITLE> '+cTitle+'</TITLE>')
         puts('</HEAD>')
         puts('')
      endwith

      return true




   ////// Method: StreamBody ///////////////////////////////////////////////////
   //////         Streams out the simplest possible response content to ////////
   //////         send back to your user  //////////////////////////////////////
   ////// Param   cText - short string that can be passed as minimal response //

   /* Note:  You should -always- subclass CgiSession so that you can customize
  				your own response page except in the case of single-line response.

            Usage:

  //          		class MyWebClass of CGISession from "Webclass.cc"
               		Function StreamBody
                 		   this.fOut>puts('/// stream your own HTML here!')
               		return true
            		endClass
   */

   Function StreamBody(cText)
      cText = iif(empty(cText), 'Operation complete. Thank you!',cText)

      with (this.fOut)
         puts('   <BODY>')
         puts(cText)
      endwith

      return true




  ////// Method: StreamFooter /////////////////////////////////////////////////
  //////         Streams out closing tags for header/title block for //////////
  //////         HTML Response page sent back to user//////////////////////////

   Function streamFooter

      with (this.fOut)
         this.fOut.puts(' ')
         this.fOut.puts('   </BODY>')
         this.fOut.puts(' ')
         this.fOut.puts('</HTML>')
      endwith
      return true




  ////// Method: PassDataThrough ///////////////////////////////////////////////
  ////// 		  Passes through all received CGI data to the next form /////////

  /* Note: This method is used to "chain" Web pages together. For example,
     let's assume that you have an order page with a product lookup. Web pages
     can't go "back" unless the user clicks the Back button on the browser.
     So you -emulate- Windows-style behavior by calling the Product Lookup,
     then  a -new- copy of the invoice form with the lookup product code
     defaulted in.

     To get the data from the first Invoice Form, through the lookup form to
     the next version of the Invoice form, you have to pass along -every single
     bit of data that came in from the first form - to the Product Lookup form
     and then as controls to the final Invoice form.

     This method creates "hiddens" in the current HTML response page for all
     valid elements of this array. That data will then be sent automatically
     by the Web Browser to the next Web applet called from the page that
     carries the "pass through data". This method automatically generates the
     "hidden" HTML code required to pass a whole form of data to the next...
     and the next... etc.

     Usage: Call this method any time after streamBody() and before
     			streamFooter()

  */

  Function PassDataThrough

     cKey  = this.firstKey

     for n = 1 to this.count()

        this.fOut.puts('<INPUT TYPE="HIDDEN" NAME="'+cKey+'" VALUE="'+this[cKey]+'">')

        cKey = this.nextKey(cKey)

     next



  ////////////////////////// Database Classes //////////////////////////////////

   ////// Method: 	LoadArrayFromFields //////////////////////////////////
   ////// Purpose:	Loads existing assoc array with data from fields ///
   //////  				in a rowset passed as a param //////////////////////
   ////// Params:	   rowsetFields = query.rowset.fields object ref //////


   /* Note: All data methods in this class are based on name-matching.
   			The HTML component name that sent the data must be an
            identical match to the name of the field, including CASE.
            In other words, name your city Entryfield CITY if the field
            in your table is named CITY! If you do this unfailingly, you
            can totally automate CGI reads into table fields and
            table fields out to HTML response pages.
   */

   Function LoadArrayFromFields(rowsetFields)

      private rf, cKey
      local n

      rf = rowsetFields  // make "private" version of param

      cKey = this.FirstKey

      for n = 1 to this.count() // traverse this array

         // see if there's a field that matches the array index key
         // and see if it's empty or not.

         // if field exists
         if type('rf[cKey]') = 'O'

            // if field is not empty
            if empty(rf[cKey].value)

               this[cKey] = ""

            else

               // figure out its type and store it as value to array
               if rf[cKey].type = 'AUTOINC' or ;
                  rf[cKey].type = 'NUMERIC'  or ;
                  rf[cKey].type = 'LONG' or ;
                  rf[cKey].type = 'INTEGER'  or ;
                  rf[cKey].type = 'DOUBLE' or ;
                  rf[cKey].type = 'FLOAT'

                  this[cKey] = str(rf[cKey].value,;
                                   rf[cKey].length,;
                                   rf[cKey].decimalLength)

               elseif rf[cKey].type = 'DATE'
                   this[cKey] = dtoc(value)

               elseif rf[cKey].type = 'LOGICAL'
                   this[cKey] = iif(value,'TRUE','FALSE')

               elseif rf[cKey].type = 'BINARY' or ;
                      rf[cKey].type = 'OLE'
                   // do nothing

               else // char or memo
                   this[cKey] = rf[cKey].value

               endif

            endif

         endif

         cKey = this.nextKey(cKey)

      next

      return true



   ////// Method 		LoadFieldsFromArray ///////////////////////////////////////
   ////// Purpose: 	load data from assoc array into rowset fields /////////////
   ////// Params: 	rowset.fields /////////////////////////////////////////////

   /*      Note: This method may be called over and over again for any number
            of queries or rows within a query. It is, however, indiscriminate.
            If you run this for more than one rowset, it will attempt to
            update each and every matching field each time you run it.
            This may present a problem if you have two tables and you only
            want to update certain fields in one table and certain fields
            in the other.

            The field-related methods of this Web Class assume either:

      		1.) Only one table is in use or
            2.) Each table has unique field names or
            3.) It's OK if more than one table has the same
            	 field name and gets the same data.

            If you need to exercize care in opdating fields in one rowset
            but not in another, (for example customer->Custno and not
            invoice->Custno), we recommend you use the subclass:
            WebDMDClass.cc,, which lets you indicate both query and fieldname,
            not just fieldname as this method does. This method should
            accomodate most circumstances.
    */

	Function loadFieldsFromArray(rowsetFields,bAppend)

      bAppend = iif(empty(bAppend),false,true)

      rf = rowsetFields  // Create private instance of param

      ////// If bAppend, make new row
      if bAppend
         rf.parent.beginAppend()
      endif

      ////// Establish first key
      cKey = this.firstKey

      ///// Traverse the array
      for n = 1 to this.count()

         if type('rf[cKey]') = 'O'  // if field exists

            with (rf[ckey])         // with (field)

               if type  = "LOGICAL"   // convert text to type and store

                  value = iif(upper(this[cKey]) ='TRUE' or ;
                           upper(this[cKey]) = 'Y',true,false)

               elseif type = "NUMERIC" or ;
                  type = "INTEGER" or ;
                  type = "DOUBLE" or ;
                  type = "FLOAT" or ;
                  type = "LONG"

                  value = val(this[cKey])

               elseif type = "DATE"

                  value = ctod(this[cKey])

               elseif type = "AUTOINC" or ;
                      type = 'TIMESTAMP' or ;
                      type = 'BINARY' or ;
                      type = 'OLE'

                    // do nothing!! Autoincrement is automatic
                    // and the others not supported in HTML

               else // Char or memo

                  value = this[cKey]

               endif

            endwith

         endif
         ////// Go to next key
         cKey = this.nextKey(cKey)

      next

      if not rf.parent.endOfSet // if not end of set
         rf.parent.save()       // save row
      endif

      return true



   ////// Method: loadDatamoduleFromArray ////////////////////////////
   //////         loads data in this array into the rowset.fields ////
   //////         of the datamodule
   ////// Param:	oDataMod - a reference to a DataModule object ///////
   ////// 			bAppend - create new row?

   Function loadDataModuleFromArray(oDataMod,bAppend)

      if empty(oDataMod)
         return false // no datamodule to work with
      endif

      oD = oDataMod  // "Privatize param"  for use in macros

                     // ensure default of bAppand
      bAppend = iif(empty(bAppend),false,bAppend)

      ////// This array will track which queries have already
      ////// been appended so we don't create extra rows  or
      ////// save rows unneccessarily

      aAppends = new array()

      ////// Traverse this array of field/value pairs and update
      ////// rowset fields with data

      cKey = this.firstKey  // init first key


      For n = 1 to this.Count()  // traverse entire array

         if at('*@',cKey) = 0    // if not a datamod pair
            cKey = this.nextKey(cKey)
            loop                 // do next one
         endif

         ////// extract query name
         cQuery = substr(cKey,1,at("*@",cKey)-1)


         ////// Prepare macro to test query
         cStr = 'oD.'+cQuery

         ////// See if query exists
         Try
            oQuery = &cStr.
         catch (exception e)
            // this query does not exist!

            cKey = this.nextKey(cKey)
            loop
         endTry


         if type('oQuery.rowset') # 'O'
            // Rowset not valid???
            cKey = this.nextKey(cKey)
            loop
         endif

         ////// See if this query has been appended or edited
         ////// if not, append row if bAppend = true
         if aAppends.scan(upper(cQuery))  = 0

            aAppends.add(upper(cQuery)) // add to query array

            if bAppend
               oQuery.rowset.beginAppend()
            endif

         endif

         ////// extract field name
         cField = substr(cKey,at("*@",cKey)+2)

         if type('oQuery.rowset.fields[cField]') # 'O' // if no such field exists

            // loop and try the next
            cKey = this.nextKey(cKey)
            loop
         endif

         with (oQuery.rowset.fields[cField])       // with (field)

            ////// Convert from string to type and update field value

            if type 	= "LOGICAL"

               value  = ;
                  iif(upper(this[cKey]) = 'TRUE' or upper(this[cKey]) = 'Y',;
                  true,false)

            elseif type = "NUMERIC" or ;
               type = "INTEGER" or ;
               type = "LONG" or ;
               type = "DOUBLE" or ;
               type = "FLOAT"

               value = val(this[cKey])

            elseif type = "DATE"

               value = ctod(this[cKey])

            elseif type = 'AUTOINC' or ;
                   type = 'TIMESTAMP'  or ;
                   type = 'BINARY' or ;
                   type = 'OLE'

               	// do nothing, not supported in HTML


            else 	// char or memo

               value = this[cKey]

            endif

         endwith

         cKey = this.nextKey(cKey)

      Next

      ////// Traverse all queries and save rows
      for n = 1 to alen(aAppends,1)
         cQuery = aAppends[n]

         cStr = 'oD.'+cQuery  // Prepare string for macro

         try
            oQuery = &cStr.   // Get reference for this query
         catch (exception e)
            loop              // if fail, go to next
         endTry

         if not oQuery.rowset.endOfSet
            oQuery.rowset.save() // save row.
         endif

      next


EndClass

